@@ -35,6 +35,8 @@
 #include <QMimeType>
 #include <QMimeDatabase>
 #include <QDateTime>
+#include <QScopedPointer>
+#include <QScopeGuard>
 
 #include <kuser.h>
 #include <kmessagebox.h>
@@ -58,13 +60,13 @@ using namespace std::filesystem;
 // TODO: investigate what size we should have and consider changing.
 // this seems too large...
 // from the RFC:
-//   The maximum size of a packet is in practise determined by the client
+//   The maximum size of a packet is in practice determined by the client
 //   (the maximum size of read or write requests that it sends, plus a few
 //   bytes of packet overhead).  All servers SHOULD support packets of at
 //   least 34000 bytes (where the packet size refers to the full length,
 //   including the header above).  This should allow for reads and writes of
 //   at most 32768 bytes.
-// In practise that means we can assume that the server supports 32kb,
+// In practice that means we can assume that the server supports 32kb,
 // it may be more or it could be less. Since there's not really a system in place to
 // figure out the maximum (and at least openssh arbitrarily resets the entire
 // session if it finds a packet that is too large
@@ -75,6 +77,16 @@ using namespace std::filesystem;
 
 #define KSFTP_ISDIR(sb) (sb->type == SSH_FILEXFER_TYPE_DIRECTORY)
 
+// sftp_attributes must be freed. Use this ScopedPtr to ensure they always are!
+struct ScopedPointerCustomDeleter
+{
+    static inline void cleanup(sftp_attributes attr)
+    {
+        sftp_attributes_free(attr);
+    }
+};
+typedef QScopedPointer<sftp_attributes_struct, ScopedPointerCustomDeleter> SFTPAttributesPtr;
+
 using namespace KIO;
 extern "C"
 {
@@ -232,7 +244,7 @@ int SFTPInternal::auth_callback(const char *prompt, char *buf, size_t len,
     qCDebug(KIO_SFTP_LOG) << "Entering authentication callback, prompt=" << mPublicKeyAuthInfo->prompt;
 
     if (q->openPasswordDialogV2(*mPublicKeyAuthInfo, errMsg) != 0) {
-        qCDebug(KIO_SFTP_LOG) << "User canceled public key passpharse dialog";
+        qCDebug(KIO_SFTP_LOG) << "User canceled public key password dialog";
         return -1;
     }
 
@@ -429,7 +441,7 @@ bool SFTPInternal::createUDSEntry(const QString &filename, const QByteArray &pat
     Q_ASSERT(entry.count() == 0);
     entry.reserve(10);
 
-    sftp_attributes sb = sftp_lstat(mSftp, path.constData());
+    SFTPAttributesPtr sb(sftp_lstat(mSftp, path.constData()));
     if (sb == nullptr) {
         qCDebug(KIO_SFTP_LOG) << "Failed to stat" << path << sftp_get_error(mSftp);
         return false;
@@ -441,7 +453,6 @@ bool SFTPInternal::createUDSEntry(const QString &filename, const QByteArray &pat
         link = sftp_readlink(mSftp, path.constData());
         if (link == nullptr) {
             qCDebug(KIO_SFTP_LOG) << "Failed to readlink despite this being a link!" << path << sftp_get_error(mSftp);
-            sftp_attributes_free(sb);
             return false;
         }
         entry.fastInsert(KIO::UDSEntry::UDS_LINK_DEST, QFile::decodeName(link));
@@ -452,8 +463,7 @@ bool SFTPInternal::createUDSEntry(const QString &filename, const QByteArray &pat
             if (sb2 == nullptr) {
                 isBrokenLink = true;
             } else {
-                sftp_attributes_free(sb);
-                sb = sb2;
+                sb.reset(sb2);
             }
         }
     }
@@ -513,8 +523,6 @@ bool SFTPInternal::createUDSEntry(const QString &filename, const QByteArray &pat
         }
     }
 
-    sftp_attributes_free(sb);
-
     return true;
 }
 
@@ -1277,6 +1285,8 @@ Result SFTPInternal::openConnection()
 
     info.password.fill('x');
     info.password.clear();
+
+    return Result::pass();
 }
 #endif // 0.8.0
 
@@ -1342,18 +1352,16 @@ Result SFTPInternal::open(const QUrl &url, QIODevice::OpenMode mode) {
     const QString path = url.path();
     const QByteArray path_c = path.toUtf8();
 
-    sftp_attributes sb = sftp_lstat(mSftp, path_c.constData());
+    SFTPAttributesPtr sb(sftp_lstat(mSftp, path_c.constData()));
     if (sb == nullptr) {
         return reportError(url, sftp_get_error(mSftp));
     }
 
     switch (sb->type) {
     case SSH_FILEXFER_TYPE_DIRECTORY:
-        sftp_attributes_free(sb);
         return Result::fail(KIO::ERR_IS_DIRECTORY, url.toDisplayString());
     case SSH_FILEXFER_TYPE_SPECIAL:
     case SSH_FILEXFER_TYPE_UNKNOWN:
-        sftp_attributes_free(sb);
         return Result::fail(KIO::ERR_CANNOT_OPEN_FOR_READING, url.toDisplayString());
     case SSH_FILEXFER_TYPE_SYMLINK:
     case SSH_FILEXFER_TYPE_REGULAR:
@@ -1361,7 +1369,6 @@ Result SFTPInternal::open(const QUrl &url, QIODevice::OpenMode mode) {
     }
 
     KIO::filesize_t fileSize = sb->size;
-    sftp_attributes_free(sb);
 
     int flags = 0;
 
@@ -1436,7 +1443,10 @@ Result SFTPInternal::read(KIO::filesize_t bytes)
     Q_ASSERT(bytesRead <= static_cast<ssize_t>(bytes));
 
     if (bytesRead < 0) {
-        qCDebug(KIO_SFTP_LOG) << "Could not read " << mOpenUrl;
+        qCDebug(KIO_SFTP_LOG) << "Could not read" << mOpenUrl
+                              << sftp_get_error(mSftp)
+                              << ssh_get_error_code(mSession)
+                              << ssh_get_error(mSession);
         close();
         return Result::fail(KIO::ERR_CANNOT_READ, mOpenUrl.toDisplayString());
     }
@@ -1487,15 +1497,14 @@ Result SFTPInternal::truncate(KIO::filesize_t length)
     Q_ASSERT(mOpenFile);
 
     int errorCode = KJob::NoError;
-    sftp_attributes attr = sftp_fstat(mOpenFile);
+    SFTPAttributesPtr attr(sftp_fstat(mOpenFile));
     if (attr) {
         attr->size = length;
-        if (sftp_setstat(mSftp, mOpenUrl.path().toUtf8().constData(), attr) == 0) {
+        if (sftp_setstat(mSftp, mOpenUrl.path().toUtf8().constData(), attr.data()) == 0) {
             q->truncated(length);
         } else {
             errorCode = toKIOError(sftp_get_error(mSftp));
         }
-        sftp_attributes_free(attr);
     } else {
         errorCode = toKIOError(sftp_get_error(mSftp));
     }
@@ -1541,14 +1550,13 @@ Result SFTPInternal::sftpGet(const QUrl &url, KIO::fileoffset_t offset, int fd)
     KIO::filesize_t totalbytesread  = 0;
     QByteArray filedata;
 
-    sftp_attributes sb = sftp_lstat(mSftp, path.constData());
+    SFTPAttributesPtr sb(sftp_lstat(mSftp, path.constData()));
     if (sb == nullptr) {
         return Result::fail(toKIOError(sftp_get_error(mSftp)), url.toString());
     }
 
     switch (sb->type) {
     case SSH_FILEXFER_TYPE_DIRECTORY:
-        sftp_attributes_free(sb);
         return Result::fail(KIO::ERR_IS_DIRECTORY, url.toString());
     case SSH_FILEXFER_TYPE_SPECIAL:
     case SSH_FILEXFER_TYPE_UNKNOWN:
@@ -1560,8 +1568,8 @@ Result SFTPInternal::sftpGet(const QUrl &url, KIO::fileoffset_t offset, int fd)
 
     // Open file
     file = sftp_open(mSftp, path.constData(), O_RDONLY, 0);
+    const auto fileFree = qScopeGuard([file]{ sftp_close(file); });
     if (file == nullptr) {
-        sftp_attributes_free(sb);
         return Result::fail(KIO::ERR_CANNOT_OPEN_FOR_READING, url.toString());
     }
 
@@ -1608,7 +1616,7 @@ Result SFTPInternal::sftpGet(const QUrl &url, KIO::fileoffset_t offset, int fd)
     }
 
     bytesread = 0;
-    SFTPInternal::GetRequest request(file, sb);
+    SFTPInternal::GetRequest request(file, sb->size);
 
     for (;;) {
         // Enqueue get requests
@@ -1677,7 +1685,7 @@ Result SFTPInternal::sftpPut(const QUrl &url, int permissions, JobFlags flags, i
     uid_t owner = 0;
     gid_t group = 0;
 
-    sftp_attributes sb = sftp_lstat(mSftp, dest_orig_c.constData());
+    SFTPAttributesPtr sb(sftp_lstat(mSftp, dest_orig_c.constData()));
     const bool bOrigExists = (sb != nullptr);
     bool bPartExists = false;
     const bool bMarkPartial = q->configValue(QStringLiteral("MarkPartial"), true);
@@ -1690,7 +1698,7 @@ Result SFTPInternal::sftpPut(const QUrl &url, int permissions, JobFlags flags, i
     }
 
     if (bMarkPartial) {
-        sftp_attributes sbPart = sftp_lstat(mSftp, dest_part_c.constData());
+        SFTPAttributesPtr sbPart(sftp_lstat(mSftp, dest_part_c.constData()));
         bPartExists = (sbPart != nullptr);
 
         if (bPartExists && !(flags & KIO::Resume) && !(flags & KIO::Overwrite) &&
@@ -1708,20 +1716,16 @@ Result SFTPInternal::sftpPut(const QUrl &url, int permissions, JobFlags flags, i
                 KIO::filesize_t pos = QT_LSEEK(fd, sbPart->size, SEEK_SET);
                 if (pos != sbPart->size) {
                     qCDebug(KIO_SFTP_LOG) << "Failed to seek to" << sbPart->size << "bytes in source file. Reason given:" << strerror(errno);
-                    sftp_attributes_free(sb);
-                    sftp_attributes_free(sbPart);
                     return Result::fail(ERR_CANNOT_SEEK, url.toString());
                 }
                 flags |= KIO::Resume;
             }
             qCDebug(KIO_SFTP_LOG) << "Resuming at" << sbPart->size;
-            sftp_attributes_free(sbPart);
         }
     }
 
     if (bOrigExists && !(flags & KIO::Overwrite) && !(flags & KIO::Resume)) {
         const int error = KSFTP_ISDIR(sb) ? KIO::ERR_DIR_ALREADY_EXIST : KIO::ERR_FILE_ALREADY_EXIST;
-        sftp_attributes_free(sb);
         return Result::fail(error, url.toString());
     }
 
@@ -1767,16 +1771,13 @@ Result SFTPInternal::sftpPut(const QUrl &url, int permissions, JobFlags flags, i
                 } // bMarkPartial
 
                 if ((flags & KIO::Resume)) {
-                    sftp_attributes fstat;
-
                     qCDebug(KIO_SFTP_LOG) << "Trying to append: " << dest;
                     file = sftp_open(mSftp, dest.constData(), O_RDWR, 0);  // append if resuming
                     if (file) {
-                        fstat = sftp_fstat(file);
+                        SFTPAttributesPtr fstat(sftp_fstat(file));
                         if (fstat) {
                             sftp_seek64(file, fstat->size); // Seek to end TODO
                             totalBytesSent += fstat->size;
-                            sftp_attributes_free(fstat);
                         }
                     }
                 } else {
@@ -1833,7 +1834,6 @@ Result SFTPInternal::sftpPut(const QUrl &url, int permissions, JobFlags flags, i
             }
         } // result
     } while (result > 0);
-    sftp_attributes_free(sb);
 
     // An error occurred deal with it.
     if (result < 0) {
@@ -1842,14 +1842,13 @@ Result SFTPInternal::sftpPut(const QUrl &url, int permissions, JobFlags flags, i
         if (file != nullptr) {
             sftp_close(file);
 
-            sftp_attributes attr = sftp_stat(mSftp, dest.constData());
+            SFTPAttributesPtr attr(sftp_stat(mSftp, dest.constData()));
             if (bMarkPartial && attr != nullptr) {
                 size_t size = q->configValue(QStringLiteral("MinimumKeepSize"), DEFAULT_MINIMUM_KEEP_SIZE);
                 if (attr->size < size) {
                     sftp_unlink(mSftp, dest.constData());
                 }
             }
-            sftp_attributes_free(attr);
         }
 
         return errorCode == KJob::NoError ? Result::pass() : Result::fail(errorCode, url.toString());
@@ -1904,7 +1903,7 @@ Result SFTPInternal::sftpPut(const QUrl &url, int permissions, JobFlags flags, i
         if (dt.isValid()) {
             struct timeval times[2];
 
-            sftp_attributes attr = sftp_lstat(mSftp, dest_orig_c.constData());
+            SFTPAttributesPtr attr(sftp_lstat(mSftp, dest_orig_c.constData()));
             if (attr != nullptr) {
                 times[0].tv_sec = attr->atime; //// access time, unchanged
                 times[1].tv_sec =  dt.toSecsSinceEpoch(); // modification time
@@ -1915,7 +1914,6 @@ Result SFTPInternal::sftpPut(const QUrl &url, int permissions, JobFlags flags, i
                 if (result < 0) {
                     qCWarning(KIO_SFTP_LOG) << "Failed to set mtime for" << dest_orig;
                 }
-                sftp_attributes_free(attr);
             }
         }
     }
@@ -2223,7 +2221,7 @@ Result SFTPInternal::listDir(const QUrl &url)
         return reportError(url, sftp_get_error(mSftp));
     }
 
-    sftp_attributes dirent = nullptr;
+    SFTPAttributesPtr dirent(nullptr);
     const QString sDetails = q->metaData(QLatin1String("details"));
     const int details = sDetails.isEmpty() ? 2 : sDetails.toInt();
     UDSEntry entry;
@@ -2237,7 +2235,7 @@ Result SFTPInternal::listDir(const QUrl &url)
         long long fileType = QT_STAT_REG;
         long long size = 0LL;
 
-        dirent = sftp_readdir(mSftp, dp);
+        dirent.reset(sftp_readdir(mSftp, dp));
         if (dirent == nullptr) {
             break;
         }
@@ -2251,7 +2249,6 @@ Result SFTPInternal::listDir(const QUrl &url)
 
             link = sftp_readlink(mSftp, file.constData());
             if (link == nullptr) {
-                sftp_attributes_free(dirent);
                 return Result::fail(KIO::ERR_INTERNAL, i18n("Could not read link: %1", QString::fromUtf8(file)));
             }
             entry.fastInsert(KIO::UDSEntry::UDS_LINK_DEST, QFile::decodeName(link));
@@ -2262,8 +2259,7 @@ Result SFTPInternal::listDir(const QUrl &url)
                 if (sb == nullptr) {
                     isBrokenLink = true;
                 } else {
-                    sftp_attributes_free(dirent);
-                    dirent = sb;
+                    dirent.reset(sb);
                 }
             }
         }
@@ -2318,7 +2314,6 @@ Result SFTPInternal::listDir(const QUrl &url)
             entry.fastInsert(KIO::UDSEntry::UDS_CREATION_TIME, dirent->createtime);
         }
 
-        sftp_attributes_free(dirent);
         q->listEntry(entry);
     } // for ever
     sftp_closedir(dp);
@@ -2347,10 +2342,9 @@ Result SFTPInternal::mkdir(const QUrl &url, int permissions)
     }
 
     qCDebug(KIO_SFTP_LOG) << "Trying to create directory: " << path;
-    sftp_attributes sb = sftp_lstat(mSftp, path_c.constData());
+    SFTPAttributesPtr sb(sftp_lstat(mSftp, path_c.constData()));
     if (sb == nullptr) {
         if (sftp_mkdir(mSftp, path_c.constData(), 0777) < 0) {
-            sftp_attributes_free(sb);
             return reportError(url, sftp_get_error(mSftp));
         }
 
@@ -2362,12 +2356,10 @@ Result SFTPInternal::mkdir(const QUrl &url, int permissions)
             }
         }
 
-        sftp_attributes_free(sb);
         return Result::pass();
     }
 
     auto err = KSFTP_ISDIR(sb) ? KIO::ERR_DIR_ALREADY_EXIST : KIO::ERR_FILE_ALREADY_EXIST;
-    sftp_attributes_free(sb);
     return Result::fail(err, path);
 }
 
@@ -2383,11 +2375,10 @@ Result SFTPInternal::rename(const QUrl &src, const QUrl &dest, KIO::JobFlags fla
     QByteArray qsrc = src.path().toUtf8();
     QByteArray qdest = dest.path().toUtf8();
 
-    sftp_attributes sb = sftp_lstat(mSftp, qdest.constData());
+    SFTPAttributesPtr sb(sftp_lstat(mSftp, qdest.constData()));
     if (sb != nullptr) {
         const bool isDir = KSFTP_ISDIR(sb);
         if (!(flags & KIO::Overwrite)) {
-            sftp_attributes_free(sb);
             return Result::fail(isDir ? KIO::ERR_DIR_ALREADY_EXIST : KIO::ERR_FILE_ALREADY_EXIST, dest.url());
         }
 
@@ -2402,7 +2393,6 @@ Result SFTPInternal::rename(const QUrl &src, const QUrl &dest, KIO::JobFlags fla
             }
         }
     }
-    sftp_attributes_free(sb);
 
     if (sftp_rename(mSftp, qsrc.constData(), qdest.constData()) < 0) {
         return reportError(dest, sftp_get_error(mSftp));
@@ -2428,7 +2418,7 @@ Result SFTPInternal::symlink(const QString &target, const QUrl &dest, KIO::JobFl
     bool failed = false;
     if (sftp_symlink(mSftp, t.constData(), d.constData()) < 0) {
         if (flags == KIO::Overwrite) {
-            sftp_attributes sb = sftp_lstat(mSftp, d.constData());
+            SFTPAttributesPtr sb(sftp_lstat(mSftp, d.constData()));
             if (sb == nullptr) {
                 failed = true;
             } else {
@@ -2440,7 +2430,6 @@ Result SFTPInternal::symlink(const QString &target, const QUrl &dest, KIO::JobFl
                     }
                 }
             }
-            sftp_attributes_free(sb);
         }
     }
 
@@ -2498,8 +2487,10 @@ void SFTPInternal::slave_status() {
     q->slaveStatus((mConnected ? mHost : QString()), mConnected);
 }
 
-SFTPInternal::GetRequest::GetRequest(sftp_file file, sftp_attributes sb, ushort maxPendingRequests)
-    :mFile(file), mSb(sb), mMaxPendingRequests(maxPendingRequests)
+SFTPInternal::GetRequest::GetRequest(sftp_file file, uint64_t size, ushort maxPendingRequests)
+    : m_file(file)
+    , m_size(size)
+    , m_maxPendingRequests(maxPendingRequests)
 {
 }
 
@@ -2509,21 +2500,21 @@ bool SFTPInternal::GetRequest::enqueueChunks()
 
     qCDebug(KIO_SFTP_TRACE_LOG) << "enqueueChunks";
 
-    while (pendingRequests.count() < mMaxPendingRequests) {
+    while (m_pendingRequests.count() < m_maxPendingRequests) {
         request.expectedLength = MAX_XFER_BUF_SIZE;
-        request.startOffset = mFile->offset;
-        request.id = sftp_async_read_begin(mFile, request.expectedLength);
+        request.startOffset = m_file->offset;
+        request.id = sftp_async_read_begin(m_file, request.expectedLength);
         if (request.id < 0) {
-            if (pendingRequests.isEmpty()) {
+            if (m_pendingRequests.isEmpty()) {
                 return false;
             } else {
                 break;
             }
         }
 
-        pendingRequests.enqueue(request);
+        m_pendingRequests.enqueue(request);
 
-        if (mFile->offset >= mSb->size) {
+        if (m_file->offset >= m_size) {
             // Do not add any more chunks if the offset is larger than the given file size.
             // However this is done after adding a request as the remote file size may
             // have changed in the meantime.
@@ -2531,19 +2522,19 @@ bool SFTPInternal::GetRequest::enqueueChunks()
         }
     }
 
-    qCDebug(KIO_SFTP_TRACE_LOG) << "enqueueChunks done" << QString::number(pendingRequests.size());
+    qCDebug(KIO_SFTP_TRACE_LOG) << "enqueueChunks done" << QString::number(m_pendingRequests.size());
 
     return true;
 }
 
 int SFTPInternal::GetRequest::readChunks(QByteArray &data)
 {
-
     int totalRead = 0;
     ssize_t bytesread = 0;
+    const uint64_t initialOffset = m_file->offset;
 
-    while (!pendingRequests.isEmpty()) {
-        SFTPInternal::GetRequest::Request &request = pendingRequests.head();
+    while (!m_pendingRequests.isEmpty()) {
+        SFTPInternal::GetRequest::Request &request = m_pendingRequests.head();
         int dataSize = data.size() + request.expectedLength;
 
         data.resize(dataSize);
@@ -2553,7 +2544,7 @@ int SFTPInternal::GetRequest::readChunks(QByteArray &data)
             break;
         }
 
-        bytesread = sftp_async_read(mFile, data.data() + totalRead, request.expectedLength, request.id);
+        bytesread = sftp_async_read(m_file, data.data() + totalRead, request.expectedLength, request.id);
 
         // qCDebug(KIO_SFTP_LOG) << "bytesread=" << QString::number(bytesread);
 
@@ -2562,7 +2553,7 @@ int SFTPInternal::GetRequest::readChunks(QByteArray &data)
             data.resize(data.size() - request.expectedLength);
 
             if (bytesread == 0) {
-                pendingRequests.dequeue(); // This frees QByteArray &data!
+                m_pendingRequests.dequeue(); // This frees QByteArray &data!
             }
 
             break;
@@ -2582,23 +2573,34 @@ int SFTPInternal::GetRequest::readChunks(QByteArray &data)
             request.expectedLength -= bytesread;
             request.startOffset += bytesread;
 
-            rc = sftp_seek64(mFile, request.startOffset);
+            rc = sftp_seek64(m_file, request.startOffset);
             if (rc < 0) {
                 // Failed to continue reading
                 return -1;
             }
 
-            request.id = sftp_async_read_begin(mFile, request.expectedLength);
+            request.id = sftp_async_read_begin(m_file, request.expectedLength);
 
             if (request.id < 0) {
                 // Failed to dispatch rerequest
                 return -1;
             }
 
+            // Move the offset back to where it was before the read.
+            // The way this works is that originally the offset is at the maximum of all pending requests,
+            // read then reduces that by the amount that it came up short, we then seek to where the short request
+            // left off and make another request for the remaining data. After that we need to move the offset
+            // back to the original value - without the reduction because we re-requested the missing data!
+            rc = sftp_seek64(m_file, initialOffset);
+            if (rc < 0) {
+                // Failed to continue reading
+                return -1;
+            }
+
             return totalRead;
         }
 
-        pendingRequests.dequeue();
+        m_pendingRequests.dequeue();
     }
 
     return totalRead;
@@ -2610,14 +2612,10 @@ SFTPInternal::GetRequest::~GetRequest()
     char buf[MAX_XFER_BUF_SIZE];
 
     // Remove pending reads to avoid memory leaks
-    while (!pendingRequests.isEmpty()) {
-        request = pendingRequests.dequeue();
-        sftp_async_read(mFile, buf, request.expectedLength, request.id);
+    while (!m_pendingRequests.isEmpty()) {
+        request = m_pendingRequests.dequeue();
+        sftp_async_read(m_file, buf, request.expectedLength, request.id);
     }
-
-    // Close channel & free attributes
-    sftp_close(mFile);
-    sftp_attributes_free(mSb);
 }
 
 void SFTPInternal::requiresUserNameRedirection()
