@@ -28,7 +28,7 @@ using namespace KIO;
 
 int SMBSlave::cache_stat(const SMBUrl &url, struct stat *st)
 {
-    int cacheStatErr;
+    int cacheStatErr = 0;
     int result = smbc_stat(url.toSmbcUrl(), st);
     if (result == 0) {
         cacheStatErr = 0;
@@ -384,11 +384,6 @@ void SMBSlave::listDir(const QUrl &kurl)
         list.clear();
     };
 
-    const auto quitLoop = [&e, &flushEntries]() {
-        flushEntries();
-        e.quit();
-    };
-
     // Since slavebase has no eventloop it wont publish results
     // on a timer, since we do not know how long our discovery
     // will take this is super meh because we may appear
@@ -419,11 +414,12 @@ void SMBSlave::listDir(const QUrl &kurl)
 
     auto maybeFinished = [&] { // finishes if all discoveries finished
         bool allFinished = true;
-        for (auto discoverer : discoverers) {
+        for (const auto &discoverer : discoverers) {
             allFinished = allFinished && discoverer->isFinished();
         }
         if (allFinished) {
-            quitLoop();
+            flushEntries();
+            e.quit();
         }
     };
 
@@ -450,14 +446,6 @@ void SMBSlave::listDir(const QUrl &kurl)
         wsd->start();
 
         qCDebug(KIO_SMB_LOG) << "Modern discovery set up.";
-
-        // Conceivably discovery could be kept from finishing by breakage in the stack.
-        // At some point we'll forcefully consider it finished to prevent broken servers and the like from
-        // holding up use of potentially working servers.
-        // NB: This must only be done for SMBURLTYPE_ENTIRE_NETWORK. Other types
-        // may take a long time to discover by necessity - e.g. listing a server with a
-        // million shares or a directory with a billion files. https://bugs.kde.org/show_bug.cgi?id=427644
-        QTimer::singleShot(16000, &e, quitLoop); // max execution time!
     }
 
     qCDebug(KIO_SMB_LOG) << "Starting discovery.";
@@ -499,7 +487,7 @@ void SMBSlave::listDir(const QUrl &kurl)
     } else {
         udsentry.fastInsert(KIO::UDSEntry::UDS_NAME, ".");
         const int statErr = browse_stat_path(m_current_url, udsentry);
-        if (statErr) {
+        if (statErr != 0) {
             if (statErr == ENOENT || statErr == ENOTDIR) {
                 reportWarning(m_current_url, statErr);
             }
